/*
 * Copyright (c) 1996, 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef GAMEENGINE_H
#define GAMEENGINE_H

#include <stdarg.h>
#include <vector>
#include <fxdefs.h>
#include <FXString.h>

#include "fxrepton.h"

enum Sound {
	Sound_Cage_Spirit,
	Sound_Died,
	Sound_Broken_Egg,
	Sound_Got_Crown,
	Sound_Got_Diamond,
	Sound_Got_Key,
	Sound_Got_Time,
	Sound_Killed_Monster,
	Sound_Transport
};

enum Status {
	Status_Idle,
	Status_Playing,
	Status_Died,
	Status_TimeOut,
	Status_EndLevel
};

enum Animation {
	Animation_None,
	Animation_Transport,
	Animation_Died
};

enum Orientation { North, South, East, West };

struct Monster {
	int xpos, ypos;
	bool dead;
	SpriteObj type;
	SpriteObj sprite;
	Orientation orientation;

	Monster(int x, int y, SpriteObj t)
		: xpos(x), ypos(y),
		  type(t), sprite(t),
		  orientation(West) { dead = false; }
};

struct Transporter {
	int orig_x, orig_y;
	int dest_x, dest_y;

	Transporter(int ox, int oy, int dx, int dy)
		: orig_x(ox), orig_y(oy),
		  dest_x(dx), dest_y(dy) {}
};

class GameEngine {
public:
	int width, height;
	FXString password;
	FXString comments;

	int paused;
	int sounds;
	int xpos, ypos;		// Repton position
	SpriteObj sprite;
	int num_diamonds;
	int capsule_time;
	int game_time;
	int score;
	bool have_key;
	Status status;

	void (*showError)(const char *fmt, ...);
	void (*playSound)(Sound snd);
	void (*showAnimation)(Animation anim);

	GameEngine();
	~GameEngine();
	void setErrorFunc(void (*errfnc)(const char *fmt, ...));
	void setSoundFunc(void (*soundfnc)(Sound snd));
	void setAnimationFunc(void (*animfnc)(Animation anim));

	SpriteObj xy(int x, int y) {
		if (x < 0 || y < 0 || x >= width || y >= height)
			return Obj_Wall;
		return ref(x, y);
	}

	void clear();
	void loadData(const char *filename, bool resetscore = true);
	bool isDataLoaded();
	Monster *findMonster(int x, int y);

	void goUp();
        void goDown();
	void goLeft();
	void goRight();
	void timer(int msecs);

private:
	SpriteObj *data;

	vector<Monster *> monsters;
	vector<Transporter *> transporters;

	SpriteObj& ref(int x, int y) {
		return data[y*width + x];
	}
	bool isempty(int x, int y) {
		switch (xy(x, y)) {
		case Obj_Empty:
		case Obj_Ground1:
		case Obj_Ground2:
			return true;
		default:
			return false;
		}
	}

	void addTransporter(int ox, int oy, int dx, int dy);
	void deleteTransporters();
	void deleteMonsters();
	Transporter *findTransporter(int x, int y);
	void addMonster(int x, int y, SpriteObj type);
	void prepareMonsters();
	bool tryFallDown(int x, int y);
	bool tryFallHorizontal(int x, int y, int xd);
	void tryFallLeftOrRight(int x, int y);
	void tryFall(int x, int y);
	bool spiritTryDirection(Monster *mp, Orientation dir, int xd, int yd);
	void updateSpirit(Monster *mp);
	double monsterTryDirection(Monster *mp, int xd, int yd);
	void updateMonster(Monster *mp);
	bool tryFungus(int x, int y, int dx, int dy);
	bool updateFungus(Monster *mp);
	void moveVerticalDirect(int yd);
	void moveHorizontalDirect(int xd);
	void gotKey();
	void gotObject(SpriteObj obj);
	void useTransporter(int x, int y);
	void moveVertical(int yd);
	void moveHorizontalMoveObj(int xd, SpriteObj obj);
	void moveHorizontal(int xd);
	void killMonster(Monster *mp);
	void cageSpirit(Monster *mp, int x, int y);
	void updateEggs();
	void updateObjects();
	void updateMonsters();
	void updateFunguses();
};

#endif // !GAMEENGINE_H
